import socket
import errno
import select
import time

from ..share import constants
from ..share.log import logger
from ..share.task import Task


class Connection:
    worker = None
    client = None

    # 待发送buf
    send_buf = None

    # 接收到buf
    recv_buf = None

    task_list = None

    _connect_expire_time = None

    def __init__(self, worker):
        self.worker = worker
        self.send_buf = bytearray()
        self.recv_buf = bytearray()
        self.task_list = []

    def fetch_task_list(self):
        task_list = self.task_list
        self.task_list = []
        return task_list

    def close(self):
        if self.client:
            try:
                self.client.close()
            except:
                logger.error('exc occur.', exc_info=True)
            finally:
                self.client = None

    def closed(self):
        """
        是否关闭
        :return:
        """
        return not self.client

    def update(self):
        """
        定时操作
        :return:
        """

        while self.worker.enable and self.closed():
            try:
                self._connect()
            except KeyboardInterrupt as e:
                # 该退出了
                return
            except:
                # 其他异常，休息后继续
                time.sleep(constants.RECONNECT_INTERVAL)
                continue
            else:
                # 连接成功
                # 设置为非阻塞
                self.client.setblocking(False)
                self._on_connected()
                break

        if not self.worker.enable:
            return

        # 能到这里，一定是连接成功了
        if self.worker.working:
            wait_timeout = 0
        else:
            # 如果不是游戏中，就阻塞在这里，可以极大提升性能
            wait_timeout = self.worker.app.conn_timeout

        wlist = [self.client] if self.send_buf else []

        ready_to_read, ready_to_write, in_error = select.select(
            [self.client], wlist, [self.client], wait_timeout
        )

        # 有事件之后，都设置为非阻塞来读取数据
        self.client.setblocking(False)

        if ready_to_read:
            self._try_to_recv()

        if ready_to_write:
            self._try_to_send()

        if in_error:
            self.close()

        if self.recv_buf:
            self.task_list.extend(self._parse_recv_buf())

    def write(self, data):
        self.send_buf.extend(data)

    def send_now(self):
        """
        立即发送数据
        :return:
        """
        self._try_to_send()

    def _connect(self):
        """
        连接
        :return:
        """

        if ':' in self.worker.app.host:
            # IPV6
            socket_type = socket.AF_INET6
        else:
            socket_type = socket.AF_INET

        client = socket.socket(socket_type, socket.SOCK_STREAM)
        client.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        client.setblocking(True)
        client.settimeout(self.worker.app.conn_timeout)

        try:
            client.connect((self.worker.app.host, self.worker.app.port))
        except Exception as e:
            client.close()
            raise e

        self.client = client

    def _parse_recv_buf(self):
        task_list = []
        offset = 0
        while offset < len(self.recv_buf):
            task = Task()
            ret = task.unpack(bytes(self.recv_buf[offset:]))
            if ret < 0:
                # 说明buf有问题
                self.recv_buf = bytearray()
                break
            elif ret == 0:
                # 没有接收完
                break
            else:
                offset += ret
                task_list.append(task)

        self.recv_buf = self.recv_buf[offset:]

        return task_list

    def _on_connected(self):
        self._ask_for_task()

    def _ask_for_task(self):
        task = Task()
        task.cmd = constants.CMD_WORKER_ASK_FOR_TASK
        task.room_id = self.worker.room_id

        self.write(task.pack())
        self.send_now()

    def _try_to_recv(self):
        while not self.closed():
            try:
                chunk = self.client.recv(self.worker.app.config['RECV_CHUNK_SIZE'])
                if not chunk:
                    self.close()
                    break

                self.recv_buf.extend(chunk)

            except socket.error as e:
                if e.errno in (errno.EINTR, errno.EAGAIN):
                    # 中断 或者 没有可读数据(非阻塞模式)
                    break
                else:
                    # Connection reset by peer 的原因说明:
                    # 网上说是对端非正常关闭连接，比如对端程序异常退出之类
                    # 我重现的方法是: C向S发送数据，如果S有回应，而C没有读取，C就调用close或者被析构的话
                    logger.error('exc occur.', exc_info=True)
                    self.close()
                    break
            except KeyboardInterrupt as e:
                # 中断
                raise e
            except:
                logger.error('exc occur.', exc_info=True)
                # 其他都直接关闭
                self.close()
                break

    def _try_to_send(self):
        """
        尝试发送，因为考虑到实时性，应该尽快发送出去
        :return:
        """

        while not self.closed() and self.send_buf:
            try:
                ret = self.client.send(self.send_buf)
                if ret < 0:
                    # 说明报错
                    break
                else:
                    self.send_buf = self.send_buf[ret:]
            except:
                logger.error('exc occur.', exc_info=True)
                break




NAME = 'pansy'

CMD_CLIENT_REQ              = 10  # 透传client请求
CMD_CLIENT_CLOSED           = 20  # 客户端连接被关闭

CMD_WRITE_TO_WORKER         = 100 # 给worker发送消息

CMD_WORKER_ASK_FOR_TASK     = 210 # 请求任务

CMD_WRITE_TO_USERS          = 230 # 主动下发
CMD_CLOSE_USERS             = 250 # 关闭多个客户端

# 重连等待时间
RECONNECT_INTERVAL = 1


DEFAULT_CONFIG = {
    'HOST': '127.0.0.1',
    'PORT': 9700,

    # 启动的room id 列表，闭区间
    'ROOM_ID_BEGIN': None,
    'ROOM_ID_END': None,

    'DEBUG': False,

    'BOX_CLASS': 'netkit.box.Box',

    'LISTENER_CLASS': None,

    'NAME': NAME,

    'CONN_TIMEOUT': 3,

    'STOP_TIMEOUT': None,

    'FPS': 10,

    # recv chunk大小
    'RECV_CHUNK_SIZE': 4096,
}



from ... import BaseListener, constants
from ...share.log import logger
from ...share.utils import safe_call
from .request import Request


class Application(BaseListener):

    box_class = None
    router = None

    def on_start(self):
        assert self.router is not None

        self.router.events.start_worker(self)

    def on_stop(self):
        self.router.events.stop_worker(self)

    def on_update(self, dt, task_list):
        dt_ms = round(dt * 1000)

        self.router.events.before_update(self, dt_ms)

        # logger.debug('dt_ms: %s, task_list: %s', dt_ms, task_list)
        if task_list:
            logger.debug('dt_ms: %s, task_list: %s', dt_ms, task_list)

            for task in task_list:
                safe_call(self._handle_task, task)

        self.router.events.after_update(self, dt_ms)

    def write_to_users(self, *args, **kwargs):
        """
        格式为
        [(uids, box), (uids, box), (uids, box) ...]
        :param data_list:
        :return:
        """
        func_name = 'write_to_users'

        self.router.events.before_response(self, (func_name, args, kwargs))
        try:
            return self.worker.write_to_users(*args, **kwargs)
        finally:
            self.router.events.after_response(self, (func_name, args, kwargs))

    def close_users(self, *args, **kwargs):
        func_name = 'close_users'

        self.router.events.before_response(self, (func_name, args, kwargs))
        try:
            return self.worker.close_users(*args, **kwargs)
        finally:
            self.router.events.after_response(self, (func_name, args, kwargs))

    def _handle_task(self, task):
        request = Request(self, task)

        if not request.is_valid:
            logger.error('invalid request. request: %s' % request)
            return

        if task.cmd == constants.CMD_CLIENT_CLOSED:
            self.router.events.close_client(request)
            return

        if not request.view_func:
            logger.error('cmd invalid. request: %s' % request)
            return False

        self.router.events.before_request(request)
        view_func_exc = None

        try:
            request.view_func(request)
        except Exception as e:
            logger.error('view_func raise exception. request: %s, e: %s',
                         request, e, exc_info=True)
            view_func_exc = e

        self.router.events.after_request(request, view_func_exc)



from .router import Router

def create_app(box_class, router):
    from .application import Application

    class MyApp(Application):
        pass

    MyApp.box_class = box_class
    MyApp.router = router

    return MyApp


from ...share.log import logger


class Request:
    """
    请求
    """

    app = None
    task = None
    box = None
    is_valid = False
    route_rule = None

    def __init__(self, app, task):
        self.app = app
        self.task = task
        self.is_valid = self._parse_raw_data()

        self.write_to_users = self.app.write_to_users
        self.close_users = self.app.close_users

    def _parse_raw_data(self):
        if not self.task.body:
            return True

        try:
            self.box = self.app.box_class()
        except Exception as e:
            logger.error('parse raw_data fail. e: %s, request: %s', e, self)
            return False

        if self.box.unpack(self.task.body) > 0:
            self._parse_route_rule()
            return True
        else:
            logger.error('unpack fail. request: %s', self)
            return False

    def _parse_route_rule(self):
        if self.cmd is None:
            return

        route_rule = self.app.router.get_route_rule(self.cmd)
        if route_rule:
            # 在app层，直接返回
            self.route_rule = route_rule
            return

    @property
    def cmd(self):
        try:
            return self.box.cmd
        except:
            return None

    @property
    def view_func(self):
        return self.route_rule['view_func'] if self.route_rule else None

    @property
    def endpoint(self):
        if not self.route_rule:
            return None

        return self.route_rule['endpoint']

    def write_to_client(self, data):
        """
        写回响应
        :param data:
        :return:
        """
        data = data or dict()

        if isinstance(data, dict):
            data.update(
                cmd=self.box.cmd,
                sn=self.box.sn,
            )

        self.app.write_to_users([
            ([self.task.uid], data)
        ])

    def __repr__(self):
        return '<%s cmd: %r, endpoint: %s, task: %r>' % (
            type(self).__name__, self.cmd, self.endpoint, self.task
        )


class BaseListener:
    worker = None

    def __init__(self, worker):
        self.worker = worker

    def on_update(self, dt, task_list):
        """
        必须继承实现
        :param dt: 等待时间
        :param task_list: 消息列表
        :return:
        """
        pass

    def on_start(self):
        """
        当worker启动后
        可继承实现
        :return:
        """
        pass

    def on_stop(self):
        """
        当worker停止前
        可继承实现
        :return:
        """
        pass



import sys

from .share.config import Config, ConfigAttribute
from .share.utils import import_module_or_string
from .share.log import logger
from .share import constants
from .master.master import Master


class Pansy:

    ############################## configurable begin ##############################
    name = ConfigAttribute('NAME')

    host = ConfigAttribute('HOST')
    port = ConfigAttribute('PORT')

    room_id_begin = ConfigAttribute('ROOM_ID_BEGIN')
    room_id_end = ConfigAttribute('ROOM_ID_END')

    box_class = ConfigAttribute('BOX_CLASS',
                                get_converter=import_module_or_string)
    listener_class = ConfigAttribute('LISTENER_CLASS',
                                     get_converter=import_module_or_string)

    debug = ConfigAttribute('DEBUG')

    conn_timeout = ConfigAttribute('CONN_TIMEOUT')

    stop_timeout = ConfigAttribute('STOP_TIMEOUT')

    fps = ConfigAttribute('FPS')

    ############################## configurable end   ##############################

    config = None

    master = None

    def __init__(self):
        self.config = Config(defaults=constants.DEFAULT_CONFIG)
        self.master = Master(self)

    def run(self):
        self.master.run()

    def make_proc_name(self, subtitle):
        """
        获取进程名称
        :param subtitle:
        :return:
        """
        proc_name = '[%s:%s %s] %s' % (
            constants.NAME,
            subtitle,
            self.name,
            ' '.join([sys.executable] + sys.argv)
        )

        return proc_name




from pyglet.clock import Clock
import signal
import setproctitle

from ..share import constants
from ..share.log import logger
from ..share.utils import safe_call
from .connection import Connection
from ..share.task import Task
from ..proto.pansy_pb2 import RspToUsers, CloseUsers


class Worker:
    app = None
    room_id = None

    enable = True

    # 是否在tick
    _working = False

    conn = None
    clock = None

    # 监听对象
    listener = None

    def __init__(self, app, room_id):
        self.app = app
        self.room_id = room_id
        self.listener = self.app.listener_class(self)

    def run(self):
        setproctitle.setproctitle(self.app.make_proc_name('worker:%s' % self.room_id))
        self._handle_signals()

        self.conn = Connection(self)

        self.on_start()

        # 房间必须有效才进行循环
        while self.enable:
            try:
                dt = 0
                # 获取网络消息放在之前dt更准一些
                if self.clock:
                    dt = self.clock.tick()
                self.update(dt)
                # logger.debug('fps: %s', clock.get_fps())
            except KeyboardInterrupt:
                break
            except:
                logger.error('exc occur.', exc_info=True)

        self.on_stop()

    def update(self, dt):
        """
        :param dt: 经过时间
        :return:
        """
        self.conn.update()
        task_list = self.conn.fetch_task_list()
        self.on_update(dt, task_list)
        # 立即发送
        self.conn.send_now()

    @property
    def working(self):
        return self._working

    @working.setter
    def working(self, value):
        self._working = value

        if self._working:
            if not self.clock:
                self.clock = Clock()
                self.clock.set_fps_limit(self.app.fps)
        else:
            if self.clock:
                self.clock = None

    def on_update(self, dt, task_list):
        """
        必须继承实现
        :param dt: 等待时间
        :param task_list: 消息列表
        :return:
        """
        safe_call(self.listener.on_update, dt, task_list)

    def on_start(self):
        """
        当worker启动后
        可继承实现
        :return:
        """
        safe_call(self.listener.on_start)

    def on_stop(self):
        """
        当worker停止前
        可继承实现
        :return:
        """
        safe_call(self.listener.on_stop)

    def write_to_users(self, data_list):
        """
        格式为
        [(uids, box), (uids, box), (uids, box) ...]
        :param data_list:
        :return:
        """

        msg = RspToUsers()

        for uids, data in data_list:
            if isinstance(data, self.app.box_class):
                data = data.pack()
            elif isinstance(data, dict):
                data = self.app.box_class(data).pack()

            row = msg.rows.add()
            row.buf = data
            row.uids.extend(uids)

        task = Task()
        task.cmd = constants.CMD_WRITE_TO_USERS
        task.body = msg.SerializeToString()

        return self.conn.write(task.pack())

    def close_users(self, uids):
        msg = CloseUsers()
        msg.uids.extend(uids)

        task = Task()
        task.cmd = constants.CMD_CLOSE_USERS
        task.body = msg.SerializeToString()

        return self.conn.write(task.pack())

    def _handle_signals(self):
        def safe_stop_handler(signum, frame):
            self.enable = False

        # 忽略掉
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # 安全停止
        signal.signal(signal.SIGTERM, safe_stop_handler)


